/*
    config.c - config parser for dchroot
    Copyright (c) 2004  Martin Schulze <joey@infodrom.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include "dchroot.h"

#define DELIM      ":;,"
#define CONF_INIT  5
#define CONF_DELTA 2

struct chroot_map *parse_line (char *line)
{
    struct chroot_map *map;
    char *name;
    char *subdir;
    char *users;
    char *p = line;
    char *bp, *cp;
    size_t len;

    /* Find the first delimiter */
    for (bp=p; *bp && !isspace(*bp) && (strchr(DELIM, *bp) == NULL); bp++);

    if (bp == p)
	return NULL;

    len = bp - p;
    if ((name = (char *)malloc((len+1)*sizeof(char *))) == NULL) {
	perror("malloc");
	exit(1);
    }
    strncpy(name, p, len);
    name[len] = '\0';

    /* Find the next non-delimiter */
    for (; *bp && (isspace(*bp) || (strchr(DELIM, *bp) != NULL)); bp++);
    p = bp;

    /* Find the next delimiter */
    for (; *bp && !isspace(*bp) && (strchr(DELIM, *bp) == NULL); bp++);

    if (bp == p) {
	free(name);
	return NULL;
    }

    len = bp - p;
    if ((subdir = (char *)malloc((len+1)*sizeof(char *))) == NULL) {
	perror("malloc");
	exit(1);
    }
    strncpy(subdir, p, len);
    subdir[len] = '\0';

    if ((map = (struct chroot_map *)malloc(sizeof(struct chroot_map *))) == NULL) {
        perror("malloc");
        exit(1);
    }
    map->name = name;
    map->subdir = subdir;
    map->users = NULL;

    /*
      "  joey  troup"
       ^
     */

    /* Find the next non-delimiter */
    for (; *bp && (isspace(*bp) || (strchr(DELIM, *bp) != NULL)); bp++);
    p = cp = bp;

    if (*p) {
	while (*bp) {
	    if (cp - p > 0)
		(*cp++) = ',';

	    /* Find the next delimiter */
	    for (; *bp && !isspace(*bp) && (strchr(DELIM, *bp) == NULL); bp++)
		(*cp++) = *bp;

	    /* Find the next non-delimiter */
	    for (; *bp && (isspace(*bp) || (strchr(DELIM, *bp) != NULL)); bp++);
	}
	/* Terminating \0 */
	(*cp++) = *bp;

	len = strlen(p);
	if ((users = (char *)malloc((len+1)*sizeof(char *))) == NULL) {
	    perror("malloc");
	    exit(1);
	}
	strncpy(users, p, len);
	users[len] = '\0';
	map->users = users;
    }

    return map;
}

struct chroot_map **parse_config(const char *conffile)
{
    FILE *cf;
    char buf[200];
    unsigned int lineno = 1;
    char *p;
    struct chroot_map *map;
    struct chroot_map **chroots = NULL;
    unsigned int chroots_len, chroots_pivot;

    if ((cf = fopen(conffile, "r")) == NULL) {
	perror("fopen");
	exit(1);
    }

    while (fgets(buf, sizeof(buf), cf) != NULL) {
	if (strlen(buf) > 0) {
	    if (buf[strlen(buf)-1] == '\n') {
		buf[strlen(buf)-1] = '\0';
	    } else {
		fprintf(stderr, "parse_line: line %d too long.\n", lineno);
		exit(1);
	    }

	    for (p=buf; *p && isspace(*p); p++);

	    if (*p == '#')
		*p = '\0';

	    if (strlen(p) > 0) {

		if ((map = parse_line (p)) == NULL) {
		    fprintf(stderr, "parse_line: line %d not parseable.\n", lineno);
		    exit(1);
		}
		if (chroots == NULL) {
		    chroots_len = CONF_INIT;
		    if ((chroots = (struct chroot_map **)malloc(chroots_len * sizeof(struct chroot_map *))) == NULL) {
			perror("malloc");
			exit(1);
		    }
		    chroots_pivot = 0;
		    memset (chroots, 0, chroots_len * sizeof(struct chroot_map *));
		} else {
		    if (!(chroots_pivot+1 < chroots_len)) {
			unsigned int i;
			if ((chroots = (struct chroot_map **)realloc(chroots,
				(chroots_len+CONF_DELTA) * sizeof(struct chroot_map *)))
			    == NULL) {
			    perror("realloc");
			    exit(1);
			}
			chroots_len += CONF_DELTA;
			for (i=chroots_pivot; i < chroots_len; i++)
			    chroots[i] = NULL;
		    }
		}
		chroots[chroots_pivot++] = map;
	    }
	}
	lineno++;
    }

    if (fclose(cf) != 0) {
	perror("fclose");
	exit(1);
    }

    return chroots;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
